from collections import namedtuple
from .route_helpers import improve_route
from .time_custom import Time_Custom
from .hash import Hash


class RouteBuilder():
    '''Class to build a single route, from hub to hub, for a truck.

    The entire "API" is the build_route method. All other methods are helpers.

    Notes on namedtuples used
    -------------------------
    A 'Neighbor' is a namedtuple, comprising:
        - loc: a Location*'s num property
        - dist: distance from previous stop

    A 'Stop' on a 'Route' is a namedtuple, comprising:
        - loc: a Location*'s num property
        - dist: distance from previous stop
        - pkgs: list of packages to drop off at this location
    A Route is then simply a list of Stops.

    A 'Stop' is upgraded to a 'StopPlus' at the end.
    A StopPlus has a Location* instead of just a location number, plus it has
        - arrival: a projected arrival time (Time_Custom objec)

    * A Location is itself a namedtuple of num, landmark, address.
    '''
    Neighbor = namedtuple('Neighbor', ['loc', 'dist'])
    Stop = namedtuple('Stop', ['loc', 'dist', 'pkgs'])
    StopPlus = namedtuple('StopPlus', ['loc', 'dist', 'pkgs', 'arrival'])

    def __init__(self, route_parameters):
        self.ready_pkgs = route_parameters['available_packages']
        self.distances = route_parameters['distances']
        self.max_load = route_parameters['max_load']
        self.truck_num = route_parameters['truck_number']
        self.Locations = route_parameters['Locations']
        self.speed_function = route_parameters['speed_function']
        self.starting_location = route_parameters['starting_location']
        self.initial_leave = route_parameters['initial_leave_time']

        self.route = []

    def get_locations(self):
        '''Return list of location-numbers/locations currently in route.'''
        return [stop.loc for stop in self.route]

    def get_packages(self):
        '''Return list of all packages currently in route.'''
        return [pkg for stop in self.route for pkg in stop.pkgs]

    def sort_by_hub_closeness(self, pkgs=None):
        '''Return packages sorted by closeness to hub.'''
        return sorted(
            pkgs if pkgs else self.get_packages(),
            key=lambda pkg: self.distances[1][pkg.props['location'].num])

    def compute_dist(self):
        '''Return total distance of route.'''
        return round(sum([stop.dist for stop in self.route]), 2)

    def display_packages(self, pkgs):
        '''Pretty-print list of packages passed in.'''
        print('\n'.join([str(pkg) for pkg in pkgs]))

    def display_stop(self, stop):
        '''Display one stop.'''
        pkgs = ',\t'.join(['Package #' + str(p.props['ID']) for p in
                           sorted(stop.pkgs, key=lambda p: p.props['ID'])])
        print(f'Stop {self.route.index(stop)+1}: going to '
              f'location #{stop.loc}, which is {stop.dist} miles from '
              f'the last stop, and has these packages:\n\t{pkgs}')

    def display_route(self, called_by='', show_load=False):
        '''Display route.'''
        if called_by:
            print(f'{called_by}')

        print('{:-^79}'.format(f' This Route is {self.compute_dist()}miles '))

        for stop in self.route:
            self.display_stop(stop)

        if show_load:
            print(f'Load has {len(self.get_packages())} pkgs, including:')
            for pkg in self.get_packages():
                print(str(pkg))

        print('*' * 79, '\n')

    def grouped_deliver_with_constraints(self):
        '''Return list of lists of ready packages, sorted smaller first,
        each list comprising packages that must be delivered together.'''
        deliver_withs = [pkg for pkg in self.ready_pkgs
                         if pkg.props['special_note']['deliver_with']]
        sets = []
        for pkg in deliver_withs:
            IDs = [pkg.props['ID']] + pkg.props['special_note']['deliver_with']
            IDs = set(IDs)
            # if this is a subset of a set in sets, skip
            if any(IDs.issubset(set_) for set_ in sets):
                continue
            # if a strict subset of this exists in sets, delete that, add this
            IDs_subsets = [set_ for set_ in sets if IDs.issuperset(set_)]
            sets = [set_ for set_ in sets if set_ not in IDs_subsets]
            sets.append(IDs)
            # if this overlaps with a set in sets, merge into this and replace
            # if there are no overlaps, neither IDs nor sets will be changed
            overlaps = [set_ for set_ in sets if not IDs.isdisjoint(set_)]
            sets = [set_ for set_ in sets if set_ not in overlaps]
            IDs = IDs.union(*overlaps)
            sets.append(IDs)

        pkgs_from_IDs = [[pkg for pkg in self.ready_pkgs
                          if pkg.props['ID'] in set_]
                         for set_ in sets]
        return sorted(pkgs_from_IDs, key=lambda lst: len(lst))

    def get_most_urgent_packages(self):
        '''Return list of packages left with deadline before 10:00 AM.'''
        return [pkg for pkg in self.ready_pkgs
                if pkg.props['deadline'] and
                pkg.props['deadline'] <= Time_Custom(10, 00, 00)]

    def get_truck_constraint_packages(self):
        '''Return list of packages left that must go on this truck.'''
        return [pkg for pkg in self.ready_pkgs
                if pkg.props['special_note']['truck_number'] and
                pkg.props['special_note']['truck_number'] == self.truck_num]

    def get_other_deadline_packages(self):
        '''Return list of packages left with deadline after 10:00 AM.'''
        return [pkg for pkg in self.ready_pkgs
                if pkg.props['deadline'] and
                pkg.props['deadline'] > Time_Custom(10, 00, 00)]

    def get_packages_on_the_way(self, pkg_load):
        '''Return list of packages destined for same place as any package in
        passed-in list (called before any packages are added to route).'''
        locations_for_pkg_load = list(set([pkg.props['location'].num
                                           for pkg in pkg_load]))
        return [pkg for pkg in list(set(self.ready_pkgs) - set(pkg_load))
                if pkg.props['location'].num in locations_for_pkg_load]

    def forbid_overfilling_load(self, so_far, more):
        '''Return package-list whose size does not exceed self.max_load, and
        which combines lists so_far and more (note: so_far could be empty).'''
        if len(so_far) + len(more) > self.max_load:
            closer_first = self.sort_by_hub_closeness(more)
            max_minus_current = self.max_load - len(so_far)
            more = closer_first[:max_minus_current]
        return list(set(so_far).union(set(more)))

    def forbid_partial_deliver_groups(self, groups, pkgs_to_load):
        '''Return package-list having no partial deliver-with groups.'''
        updated = pkgs_to_load[:]
        for pkg in pkgs_to_load:
            for deliver_group in groups:
                if pkg in deliver_group:
                    together = list(set(updated).union(set(deliver_group)))
                    if len(together) <= self.max_load:
                        updated = together
                    else:
                        updated.remove(pkg)
        return updated

    def packages_left(self):
        '''Return list of packages ready to go not already in route.'''
        return list(set(self.ready_pkgs) - set(self.get_packages()))

    def unvisited_stops_with_packages(self):
        '''Return list of unvisited locations at which at least one ready,
        unpicked package needs to be dropped off.'''
        stops_with_pkgs = list(set([
            pkg.props['location'].num for pkg in self.packages_left()]))
        # start at 2 because col 0 isn't distance data, and col 1 is the hub
        return [loc_num for loc_num in self.distances[0][2:]
                if loc_num in stops_with_pkgs and
                loc_num not in self.get_locations()]

    def find_nearest(self, Stop, location_list=None):
        '''Return nearest neighbor-with-packages to Stop passed in.'''
        starting_from = zip(self.distances[0], self.distances[Stop.loc])
        eligible_location_nums = (location_list if location_list
                                  else self.unvisited_stops_with_packages())
        eligible_neighbors = [RouteBuilder.Neighbor(loc_num, dist)
                              for (loc_num, dist) in starting_from
                              if loc_num in eligible_location_nums]
        return min(eligible_neighbors,
                   key=lambda neighbor: neighbor.dist)

    def add_first_stop(self):
        '''Add first stop.'''
        self.route.append(RouteBuilder.Stop(self.starting_location, 0, []))

    def add_final_stop(self):
        '''Add final stop.'''
        dist = self.distances[self.route[-1].loc][1]  # from previous to hub
        self.route.append(RouteBuilder.Stop(self.starting_location, dist, []))

    def get_earliest_deadline_for_stop(self, stop):
        '''Return earliest deadline, if any, for a given stop in the route.'''
        deadlines = [pkg.props['deadline'] for pkg in stop.pkgs
                     if pkg.props['deadline']]
        return min(deadlines) if len(deadlines) > 0 else None

    def Location_from_number(self, num):
        '''Return a Location from a location-number.'''
        return [L for L in self.Locations if L.num == num][0]

    def get_projected_arrival(self, stop):
        '''Return projected arrival time for a stop on a route.'''
        index = self.route.index(stop)
        if index == 0:
            return self.initial_leave
        previous = self.route[index - 1]
        avg_speed = self.speed_function(previous.loc, stop.loc)
        minutes = 60 * (stop.dist / avg_speed)
        projected_arrival = Time_Custom.clone(previous.arrival)
        projected_arrival.add_time(minutes)
        return projected_arrival

    def convert_to_stopplus(self):
        '''Replace each Stop in route with a StopPlus--add projected arrival,
        and replace location number with reference to a Location.'''
        for index, stop in enumerate(self.route):
            Location = self.Location_from_number(stop.loc)
            projected_arrival = self.get_projected_arrival(stop)
            self.route[index] = RouteBuilder.StopPlus(
                Location, stop.dist, stop.pkgs, projected_arrival)

    def add_more_deliverwith_groups(self, pkgs_to_load, groups):
        '''Add more deliver-with groups that will fit, smallest first.'''
        for group in groups:  # deliver-with method sorted them smallest-first
            new_list = list(set(pkgs_to_load).union(set(group)))
            if len(new_list) <= self.max_load:
                pkgs_to_load = new_list

        delivwith_pkgs_left = [pkg for group in groups for pkg in group]
        self.ready_pkgs = list(set(self.ready_pkgs) - set(delivwith_pkgs_left))

        return pkgs_to_load

    def construct_stops(self, pkgs_to_load):
        '''Construct stops for packages, in nearest-neighbor order, and append
        them to the route.'''
        locs = list(set([pkg.props['location'].num for pkg in pkgs_to_load]))
        while len(locs) > 0:
            nearest = self.find_nearest(self.route[-1], locs)
            pkgs_for_stop = [pkg for pkg in pkgs_to_load
                             if pkg.props['location'].num == nearest.loc]
            self.route.append(RouteBuilder.Stop(
                nearest.loc, nearest.dist, pkgs_for_stop))
            locs.remove(nearest.loc)

    def add_nearby_neighbors(self, acceptable_increase):
        '''Add nearby neighbors found throughout the route-so-far if doing so
        wouldn't increase distance much (determined by acceptable_increase).'''
        stop_index = 0

        while (len(self.get_packages()) < self.max_load and
               len(self.unvisited_stops_with_packages()) > 0):

            stop = self.route[stop_index]
            if stop == self.route[-1]:
                break

            nearest = self.find_nearest(stop)
            cur_next = self.route[stop_index + 1]
            if nearest == cur_next:
                continue

            distance_with_nearest = (self.distances[stop.loc][nearest.loc] +
                                     self.distances[nearest.loc][cur_next.loc])

            if distance_with_nearest <= acceptable_increase * cur_next.dist:
                for_here = [pkg for pkg in self.packages_left()
                            if pkg.props['location'].num == nearest.loc]
                excess_removed = self.forbid_overfilling_load(
                    self.get_packages(), for_here)
                for_here = list(set(excess_removed) - set(self.get_packages()))

                if len(for_here) > 0:
                    self.route.insert(stop_index + 1, (RouteBuilder.Stop(
                        nearest.loc, nearest.dist, for_here)))

            stop_index += 1

    def add_stops_at_end(self):
        '''Add stops at end using nearest-neighbors until load is full or
        no more destination stops exist.'''
        while (len(self.get_packages()) < self.max_load and
               len(self.unvisited_stops_with_packages()) > 0):

            nearest = self.find_nearest(self.route[-1])

            more = [pkg for pkg in self.packages_left()
                    if pkg.props['location'].num == nearest.loc]
            all_pkgs = self.forbid_overfilling_load(self.get_packages(), more)
            at_this_stop = list(set(all_pkgs) - set(self.get_packages()))

            if len(at_this_stop) > 0:
                self.route.append(RouteBuilder.Stop(
                    nearest.loc, nearest.dist, at_this_stop))

    def build_route(self):
        '''Return a delivery route (list of stops).'''
        if len(self.ready_pkgs) == 0:
            return []

        self.add_first_stop()

        groups = self.grouped_deliver_with_constraints()

        #    I.    Add urgent packages and truck-constraint packages first.
        # If truck is leaving at or after 9am, get packages with non-urgent
        # deadlines too. Also add any that must be delivered with those.
        # Note that no Stops for the route are being created yet.
        pkgs_to_load = self.get_most_urgent_packages()
        pkgs_to_load = self.forbid_overfilling_load([], pkgs_to_load)

        more_to_load = self.get_truck_constraint_packages()
        if self.initial_leave > Time_Custom(8, 00, 00):
            more_to_load += self.get_other_deadline_packages()
        pkgs_to_load = self.forbid_overfilling_load(pkgs_to_load, more_to_load)

        pkgs_to_load = self.forbid_partial_deliver_groups(groups, pkgs_to_load)

        #    II.   Get other packages that would be 'on the way'.
        # Also get any that must be delivered with those.
        more_to_load = self.get_packages_on_the_way(pkgs_to_load)
        pkgs_to_load = self.forbid_overfilling_load(pkgs_to_load, more_to_load)
        pkgs_to_load = self.forbid_partial_deliver_groups(groups, pkgs_to_load)

        #    III.  Add other deliver-with groups that will fit, smallest-first,
        # then remove packages in remaining groups from consideration. The idea
        # for the first part of that is to get deliver-with packages out of the
        # way as soon as possible, and the idea for the second part is to not
        # have to worry about partial groups again while route-building.
        #
        # Note: it is likely the first route of the day will be longer than
        # successive trips, as it is mostly driven by package constraints,
        # rather than nearest-neighbors / distance.
        pkgs_to_load = self.add_more_deliverwith_groups(pkgs_to_load, groups)

        #    IV.   Construct stops from pkgs_to_load and add to route.
        self.construct_stops(pkgs_to_load)

        #    V.    Look for nearby neighbors between each stop-pair on route
        # PLEASE NOTE: ~1.6 is simply the parameter that performed well for me
        # with my sample data. It may need to be changed if given more data!
        acceptable_increase = 1.6
        self.add_nearby_neighbors(acceptable_increase)

        #    VI.   Add more stops near the end of the route
        self.add_stops_at_end()

        #    VII.  Re-order stops on route to get shorter total distance,
        # so long as deadlines wouldn't be missed. Note that the code up to
        # now doesn't itself guarantee deadlines will be missed, but luckily
        # it is a natural place to check for deadlines in the same place
        # (namely, improve_route) we generate permutations and select the best.
        self.add_final_stop()

        stop_deadlines = [(stop.loc, self.get_earliest_deadline_for_stop(stop))
                          for stop in self.route]
        stop_deadlines = [sd for sd in stop_deadlines if sd[1]]  # remove Nones

        self.route = improve_route(self.route, self.distances, stop_deadlines,
                                   self.speed_function, self.initial_leave,
                                   RouteBuilder.Stop)

        #    VIII. Convert Stops on route to StopPluses and return route
        self.convert_to_stopplus()
        return self.route



import re
from collections import namedtuple
from enum import Enum
from .time_custom import *
from .hash import *


class PackageSpecialNote_ValueError(BaseException):
    pass


class PkgState(Enum):
    '''Enumerate possible package states.'''
    AT_HUB = 1
    IN_TRANSIT = 2
    DELIVERED = 3
    LATE_ARRIVAL = 4
    WRONG_DESTINATION = 5


class Package():
    '''The Package class provides Package objects.

    Attributes (Instance variables):
     - ID: ID
     - deadline: when a package must be delivered by
     - weight: in kilograms.
        Note: no aspect of this entire program currently cares about weight.
    - location: a namedtuple of num, landmark, address
    - special note: a property of a package that itself has 4 properties:
        - truck_number
        - deliver_with
        - late_arrival
        - wrong_destination
        The first two of these subproperties are used to set initial state.
    '''

    History_Record = namedtuple('History_Record', ['state', 'time'])

    def __init__(self, pkg_id, d, w, sn, location):
        '''Create Package object.'''
        self.props = Hash(ID=int(pkg_id),
                          deadline=d,
                          weight=w,
                          location=location)

        self.props['special_note'] = Hash('truck_number',
                                          'deliver_with',
                                          'late_arrival',
                                          'wrong_destination')
        self.mark_package_special(self.parse_special_note(sn))

        self.props['state'] = None
        self.set_initial_state()

        self.props['history'] = []
        self.set_initial_history()

    def set_state(self, state_string):
        '''Update state of a package.'''
        self.props['state'] = PkgState(PkgState[state_string])

    def set_initial_state(self):
        '''Set initial state of package to hub, late, or wrong-destination.'''
        if self.props['special_note']['late_arrival']:
            self.set_state('LATE_ARRIVAL')
        elif self.props['special_note']['wrong_destination']:
            self.set_state('WRONG_DESTINATION')
        else:
            self.set_state('AT_HUB')

    def set_initial_history(self):
        '''Set initial history of a package as at-hub at 7:59am.'''
        self.props['history'].append(
            Package.History_Record(self.props['state'],
                                   Time_Custom(7, 59, 00)))

    def add_to_history(self, state_string, time):
        '''Add to history of a package object.'''
        self.props['history'].append(
            Package.History_Record(PkgState(PkgState[state_string]), time))

    def history_string(self, delimiter=None):
        '''Return print-statement-friendly history of package.'''
        return f'\n{delimiter}'.join([' at:\t'.join((record.state.name,
                                                     str(record.time)))
                          for record in self.props['history']])

    def update_late_as_arrived(self, time):
        '''Update a late-arriving package to indicate it is now at the hub.'''
        self.set_state('AT_HUB')
        self.props['history'].append(
            Package.History_Record(PkgState(PkgState['AT_HUB']), time))

    def update_wrong_destination_as_corrected(self):
        '''Update a wrong-destination package to indicate destination is now
        known. In other words, it can be delivered.'''
        self.set_state('AT_HUB')

    def update_package_destination(self, updated_destination):
        '''Update location property of package object.'''
        self.props['location'] = updated_destination

    def parse_special_note(self, special_note):
        '''Parse the special note (if any) attached to a package.

        The meaning of a special note is determined by its match to one of the
        following (caps-insensitive) patterns:
            1. 'only', followed by 'truck', followed by an integer
                If matched, ...
            2. 'delivered with', followed by an integer or comma-separated
            list of integers
                If matched, ...
            3. 'delayed', followed by a time of the form 'h:mm (am|pm)'
            or 'hh:mm (am|pm)'
                If matched, ...
            4. 'wrong address'
                If matched, ...
        Assumption: only one pattern will be matched for any given package.
        If more than one did match, the constraint actually assigned to
        the package is dictated by the order in which they are checked.
        '''
        truck_number_constraint = truck_regex_match(special_note)
        co_delivery_constraint = delivery_regex_match(special_note)
        late_arrival_constraint = arrival_regex_match(special_note)
        wrong_destination_constraint = destination_regex_match(special_note)

        if truck_number_constraint:
            truck_num = get_truck_number(special_note)
            return 'truck_number', truck_num
        if co_delivery_constraint:
            package_list = get_packages_to_deliver_with(special_note)
            return 'deliver_with', package_list
        if late_arrival_constraint:
            if not Time_Custom.is_valid_AM_PM_time(special_note):
                raise PackageSpecialNote_ValueError(
                    'Time-like value not parseable as an AM/PM time')
            when = Time_Custom.make_time_from_string(special_note)
            return 'late_arrival', when
        if wrong_destination_constraint:
            return 'wrong_destination', True

        if special_note != '':
            raise PackageSpecialNote_ValueError(
                'Special note could not be parsed as a package constraint')
        return None

    def mark_package_special(self, parsed_note):
        '''Update package props based off result of parsing special_note.

        Sets four new properties on self.special_note, initialized to None,
        then updates the zero or one properties passed in as parsed_note.
        '''
        if parsed_note is not None:
            parsed_note_key, parsed_note_value = parsed_note
            self.props['special_note'][parsed_note_key] = parsed_note_value

    def __str__(self):
        '''Return string representation of a Package object.'''
        return '\n\t'.join([f"Package ID: {self.props['ID']}",
                            f"delivery status: {self.props['state']}",
                            f"destination: {self.props['location'].address}",
                            f"deadline: {self.props['deadline']}",
                            f"weight: {self.props['weight']}"])


'''
    The following regex functions are not part of the Package class because
    they do not need to be--they neither consume nor produce Package objects.
'''
truck_pattern = re.compile("(only)(.*?)(truck) \d", re.IGNORECASE)
delivery_pattern = re.compile("(delivered with \d+[, \d+]*)", re.IGNORECASE)
arrival_pattern = re.compile("(delayed)(.*?)\d\d?:\d\d (am|pm)", re.IGNORECASE)
destination_pattern = re.compile("wrong address", re.IGNORECASE)


def truck_regex_match(note):
    '''Return regex match for a truck-number constraint in a special note.'''
    return truck_pattern.search(note)


def delivery_regex_match(note):
    '''Return regex match for a delivery constraint in a special note.'''
    return delivery_pattern.search(note)


def arrival_regex_match(note):
    '''Return regex match for a late-arrival constraint in a special note.'''
    return arrival_pattern.search(note)


def destination_regex_match(note):
    '''Return regex match for wrong-destination constraint in special note.'''
    return destination_pattern.search(note)


def validate_truck_number(integers_found):
    '''Validate truck number found in package special note.

    This function assumes that the company's truck-number labels
    will always be less than 100.
    '''
    return len(integers_found) == 1 and integers_found[0] < 100


def get_truck_number(note):
    '''Parse and return truck number from a package special note.'''
    integers_found = [int(num) for num in filter(str.isdigit, note)]
    if not validate_truck_number(integers_found):
        raise PackageSpecialNote_ValueError(
            'Truck-number error in special note: either more than one '
            'number was found, or it was too high to be a truck number')

    truck_num = integers_found[0]
    return truck_num


def validate_package_ID_list(integers_found):
    '''Validate co-delivery package IDs found in package special note.

    This function assumes that the company's package ID numbers
    will always be less than 1000. This assumption is shared by a
    cli module function (get_package_id_from_string).
    '''
    return all([i > 0 and i < 1000 for i in integers_found])


def get_packages_to_deliver_with(note):
    '''Parse and return list of package IDs from a package special note.'''
    index_of_first_integer = re.search("\d", note, re.IGNORECASE).start()
    substring = note[index_of_first_integer:]
    integers_found = [int(num) for num in substring.split(', ')]
    if len(integers_found) == 0:  # try split by ',' instead of ', '
        integers_found = [int(num) for num in substring.split(',')]

    if not validate_package_ID_list(integers_found):
        raise PackageSpecialNote_ValueError(
            'Package-ID error in a deliver-with special note: one or more '
            'IDs found was either below zero or too high to be a package ID')

    package_IDs = integers_found
    return package_IDs



import re


def test_regexes():
    '''
    Functions to test:
    truck_regex_match
    delivery_regex_match
    arrival_regex_match
    destination_regex_match

    get_truck_number
    get_packages_to_deliver_with
    get_arrival_time

    validate_truck_number
    validate_package_ID_list
    validate_arrival_time
    '''
    pass

# Test strings and expected pass/fails:
# s = 'Can only be on truck 2'  # expect: pass
# s = 'Can only be on truck2'  # fail
# s = 'can oNlY be on trUcK 2'  # pass
# s = 'Can only be on truck 22'  # fail
# s = 'Can only be on truck 0'  # fail
# s = 'Can only be on truck 4'  # fail
# s = 'Can only be on truck -1'  # fail
# s = 'Can only be on truck'  # fail
# s = 'Delayed on flight---will not arrive to depot until 9:05 am'  # pass
# s = 'Delay--will not arrive until 9:05 am'  # pass
# s = 'Delay--will arrive 9:05 am'  # pass
# s = 'Delay--will arrive 9:05am'  # uh oh, currently won't pass--but should
# s = 'Delay--will arrive 19:00am'  # pass (yes, user shouldn't have said am!)
# s = 'Delay--will arrive 24:00pm'  # fail
# s = 'Delay--will arrive 19:65am'  # fail
# s = 'Delay--will arrive 6am'  # fail
# s = 'Delay--will arrive 9am'  # uh oh, currently won't pass--but should
# s = 'Must be delivered with 15, 19'  # pass
# s = 'Must be delivered with 15,19'  # pass
# s = 'Must be delivered with 5, 19'  # pass
# s = 'Must be delivered with 5, 3, and 19'  # fail (users must be trained!)
# s = 'Wrong address listed'  # pass
# s = 'wrong address'  # pass
# s = 'incorrect address'  # fail (users must be trained to write 'wrong'!)



from .time_custom import *
from .package import *


class Truck():
    '''This class creates Truck objects.

    Class Attributes:
     - id_counter: to ensure Trucks have unique IDs, starting at 1
     - max_packages: all trucks carry a maximum of 16 packages at once
     - average_speed: all trucks always go at 18mph (including stops)
     - first_delivery_time: all trucks initially leave the hub at 8:00 AM

    Attributes (Instance variables):
     - ID: ID
     - location: a namedtuple of num, landmark, address
     - time: current time of truck
        This only updates when trucks arrive somewhere, not every minute.
     - packages: list of packages currently on the truck
     - mileage_for_day: mileage for the day

    Trucks, when they are at the hub, are capable of seeing whether any
    late-arrival packages have arrived and updating such packages, and they
    are also capable (or rather, you can suppose their drivers are capable) of
    updating wrong-destination packages (if the correct destination is known).
    '''

    id_counter = 1
    max_packages = 16
    average_speed = 18
    starting_location = 1  # location 1 is the hub
    first_delivery_time = Time_Custom(8, 00, 00)

    def __init__(self):
        '''Create Truck object.'''
        self.props = Hash(ID=Truck.id_counter,
                          location=Truck.starting_location,
                          time=Truck.first_delivery_time,
                          packages=[],
                          mileage_for_day=0)

        Truck.id_counter += 1

    def update_late_packages(self, all_packages):
        '''Update all late-arriving packages that are now at the hub.'''
        late_arrivals = [pkg for pkg in all_packages
                         if pkg.props['special_note']['late_arrival']
                         is not None]

        for pkg in late_arrivals:
            anticipated_arrival = pkg.props['special_note']['late_arrival']

            if (self.props['time'] > anticipated_arrival and
                    pkg.props['state'].name == 'LATE_ARRIVAL'):
                # anticipated_arrival will be used in the new History_Record,
                # which means we implicitly assume late arrivals arrive
                # precisely when we were told they would (and not even later!)
                pkg.update_late_as_arrived(anticipated_arrival)

    def update_corrected_packages(self, all_packages, destination_corrections):
        '''Update all packages that had a known wrong-destination at start of day
        but which have now been corrected, i.e., which can now be delivered.'''
        wrong_destinations = [p for p in all_packages
                              if p.props['state'].name == 'WRONG_DESTINATION']

        for pkg in wrong_destinations:
            if pkg.props['ID'] in [c.pkg_id for c in destination_corrections]:
                updated_destination, = [c for c in destination_corrections
                                        if c.pkg_id == pkg.props['ID']]

                if updated_destination.location is not None:
                    pkg.update_package_destination(
                        updated_destination.location)
                    pkg.update_wrong_destination_as_corrected()

    def get_available_packages(self, all_packages, destination_corrections):
        '''Return list of packages at hub and update late or wrong-destination
        packages that have (respectively) arrived or been corrected.'''
        self.update_late_packages(all_packages)
        self.update_corrected_packages(all_packages, destination_corrections)
        at_hub = [pkg for pkg in all_packages
                  if pkg.props['state'].name == 'AT_HUB']
        ID = self.props['ID']
        can_go = [pkg for pkg in all_packages
                  if (pkg.props['special_note']['truck_number'] is None or
                      pkg.props['special_note']['truck_number'] == ID)]
        return [pkg for pkg in all_packages
                if pkg in at_hub and pkg in can_go]

    def load(self, pkg_load):
        '''Load truck with packages.'''
        self.props['packages'] = pkg_load

    def get_mileage_for_day(self):
        '''Find and return actual mileage truck has traveled today.

        This function currently just returns the instance prop mileage_for_day
        but could easily be extended in the future to account for "real life",
        for example if a truck was forced to take a detour.
        '''
        return self.props['mileage_for_day']

    def deliver(self, route):
        '''Deliver packages on truck.

        Note: this whole program implicitly assumes that trucks are able to
        deliver each package precisely on schedule.

        A 'Stop' on a 'Route' is a namedtuple, comprising:
            - loc: location of the stop
            - pkgs: list of packages to drop off at this location
            - dist: distance from previous stop
            - arrival: a Time_Custom object (projected arrival, not actual)
        '''
        for stop in route:
            self.props['location'] = stop.loc
            self.props['time'] = stop.arrival
            self.props['packages'] = list(
                set(self.props['packages']) - set(stop.pkgs))
            self.props['mileage_for_day'] += stop.dist

            for pkg in stop.pkgs:
                pkg.set_state('DELIVERED')
                pkg.add_to_history('DELIVERED', self.props['time'])

    def __str__(self):
        '''Return string representation of Truck object.'''
        package_list = '\n\t'.join([str(pk) for pk in self.props['packages']])
        return (f"Truck with ID: {self.props['ID']}; location: "
                f"{self.props['location']}; time: {str(self.props['time'])};"
                f" with these packages: \n\t{package_list}")

    @classmethod
    def speed_function(cls, location1, location2):
        '''Return average speed between two locations in miles per hour.

        This function currently just returns the class prop average_speed
        but could easily be extended in the future to account for 'real life',
        for example some roads could be faster than others.
        '''
        return cls.average_speed



# Adam Isom, Student ID #000906109
import sys
from .cli import (say_hello, ask_if_snapshot_wanted, handle_snapshot_request,
                  ask_if_package_histories_wanted, ask_if_route_display_wanted,
                  get_destination_corrections)
from .load import load_data
from .classes.time_custom import Time_Custom
from .classes.hash import Hash
from .classes.package import Package
from .classes.truck import Truck
from .classes.route_builder import RouteBuilder
from .tests.general import test


'''
Note: the number 79 is hardcoded 5 times in "display" functions and methods
    to print out a line 79 characters long (2 times here in __main__, 2 times
    in cli, and 1 time in route_builder), so if a user's console is less than
    79 characters, the print-outs won't be pretty.
'''


def all_packages_delivered(packages):
    '''Return whether all packages are delivered.'''
    return all([pkg.props['state'].name == 'DELIVERED'
                for pkg in packages])


def number_delivered(packages):
    '''Return count of number delivered.'''
    return sum([1 for pkg in packages
                if pkg.props['state'].name == 'DELIVERED'])


def display_packages_with_history(packages):
    '''Display each package and its delivery-status histories.'''
    sorted_pkgs_with_history = '\n'.join(
        [(str(pkg).replace('PkgState.', '') + '\n' + '\t' +
          pkg.history_string('\t'))
        for pkg in sorted(packages, key=lambda p: p.props['ID'])])

    print(f'Packages and their histories:\n{sorted_pkgs_with_history}\n')
    print('*' * 79, '\n')


def display_distance_traveled(total_distance):
    '''Display distance (in miles) traveled to deliver all packages.'''
    print(f'\nTotal travel distance today was {total_distance:.2f} miles.')


def is_package_delivered_and_on_time(package):
    '''Return whether the given package was delivered at all if it had no
    deadline, and whether it was delivered on time if it did have one.'''
    for record in package.props['history']:
        if record.state.name == 'DELIVERED':
            if package.props['deadline'] is None:
                return True
            if record.time <= package.props['deadline']:
                return True
    return False


def display_number_delivered_on_time(packages):
    '''Calculate and display the number (it should be all of them) of packages
    that were delivered on time.'''
    on_time_count = sum([1 for pkg in packages
                         if is_package_delivered_and_on_time(pkg)])
    print(f'\n{on_time_count} out of {len(packages)} packages '
          'were delivered on time.')


def display_distances(distances):
    '''Print distance 2D-list as a readable table.'''
    string = '\n'.join(''.join([str(item).rjust(5, ' ')
                                for item in row])
                       for row in distances)
    string = string.replace('DISTANCE BETWEEN HUBS IN MILES', '    ')
    print('DISTANCES\n', string)


def run_program(distance_csv, package_csv):
    '''Run the program!'''
    distances, Locations, packages = load_data(distance_csv, package_csv)

    say_hello()
    Destination_Corrections = get_destination_corrections(Locations)
    route_display_wanted = ask_if_route_display_wanted()
    snapshot_wanted = ask_if_snapshot_wanted()
    package_histories_wanted = ask_if_package_histories_wanted()
    print('*' * 79, '\n')

    number_of_trucks = 3
    number_of_drivers = 2
    trucks = []
    for i in range(number_of_drivers):  # trucks can't be sent without drivers
        trucks.append(Truck())

    while not all_packages_delivered(packages):
        number_delivered_before_loop = number_delivered(packages)

        for truck in trucks:

            packages_ready = truck.get_available_packages(
                packages, Destination_Corrections)

            route_parameters = Hash(
                ['available_packages', packages_ready],
                ['distances', distances],
                ['max_load', Truck.max_packages],
                ['truck_number', truck.props['ID']],
                ['Locations', Locations],
                ['speed_function', Truck.speed_function],
                ['starting_location', Truck.starting_location],
                ['initial_leave_time', Truck.first_delivery_time])
            route_builder = RouteBuilder(route_parameters)
            route = route_builder.build_route()

            if route_display_wanted and route != []:
                route_builder.display_route()

            truck.load(route_builder.get_packages())
            truck.deliver(route)

        number_delivered_after_loop = number_delivered(packages)
        if not number_delivered_after_loop > number_delivered_before_loop:
            break

    total_distance = sum([truck.props['mileage_for_day']
                          for truck in trucks])
    display_distance_traveled(total_distance)
    display_number_delivered_on_time(packages)
    print('\n')
    print('*' * 79)

    if snapshot_wanted:
        handle_snapshot_request(packages)

    if package_histories_wanted:
        display_packages_with_history(packages)


if __name__ == '__main__':
    try:
        dist_csv = sys.argv[1]
        pkg_csv = sys.argv[2]
    except IndexError:
        raise IndexError('You didn\'t provide both required csv files')

    run_program(dist_csv, pkg_csv)



import re
from os import path
from collections import namedtuple
from .classes.time_custom import Time_Custom
from .classes.package import Package


def say_hello():
    '''Say hello when the program starts and advise the user they will be asked
    what information, if any, they would like displayed.'''
    print('Hello!\nThis application simulates package delivery via truck.')
    print('\nPlease note: A "route" is a single trip, starting at the hub\n'
          'and ending at the hub; multiple delivery routes may be needed\n'
          'to deliver all your packages.\n')
    print('This program is about to ask you a few things, in this order:\n'
          '\twhether you have any destination-correction information to enter,'
          '\n\twhether you want to display each route\'s path (with packages),'
          '\n\twhether you would like to get a snapshot at the end, showing'
          '\n\tthe delivery status of each package at some time that you '
          'provide,\n\tand finally, whether you want see the history '
          'of each package at the very end.\n')


def ask_if_route_display_wanted():
  '''Return whether user wants to view each route and its stops.'''
  user_says = input('Would you like to see each route that is calculated?\n')
  return user_says.lower().strip() in ('y', 'yes')


def ask_if_snapshot_wanted():
  '''Return whether user wants a snapshot.'''
  user_says = input('Would you like to generate a snapshot at the end?\n'
                    'If so, then once the routes are calculated (and '
                    'displayed, if you wanted that),\nyou\'ll be '
                    'asked to provide a time for the snapshot.\nThat snapshot '
                    'will appear in the console here, and will also be stored '
                    'in\na new file "package_snapshot.txt" in the directory '
                    'you are running this from.\nIf you would '
                    'like a snapshot later, press "y" or "yes", then Enter.\n'
                    'Otherwise, press any other key, like Enter or "n".\n')
  return user_says.lower().strip() in ('y', 'yes')


def ask_if_package_histories_wanted():
  '''Return whether user wants to view all package histories.'''
  user_says = input('Would you like to see all packages and their histories '
                    'at the end?\n(This will appear after a snapshot, if you '
                    'asked for one.)\n')
  return user_says.lower().strip() in ('y', 'yes')


def give_user_snapshot_instructions():
    '''Advise the user to enter a time for a snapshot.'''
    print('\nAlright, time for that snapshot you asked about.\n'
          'Please enter the time on the next line.\nMake sure to enter a '
          '\'military\' time, for example 14:00 for 2:00pm.\nSeconds are '
          'optional, but make sure to use two digits for the hour.\n'
          'Examples: 08:35, 10:00, 12:01:30 (this would be just after noon)')


def validate_length(time_string):
    '''Validate that the user-requested time-string is a correct length.'''
    return len(time_string.strip()) == 5 or len(time_string.strip()) == 7


def get_time_parts_from_string(time_string):
    '''Extract and return hh, mm, ss from a string of form hh:mm or hh:mm:ss.

    hh refers to two-digit hour, mm : two-digit minute, ss : two-digit second
    '''
    hour = int(time_string[0:2])
    minute = int(time_string[3:5])
    second = (int(time_string[6:8])
              if len(time_string.strip()) == 7
              else 0)
    return hour, minute, second


def validate_values(time_parts):
    '''Validate user-provided time for snapshot has a valid number of hours,
    minutes and seconds.'''
    hour, minute, second = time_parts
    return (hour >= 0 and hour < 24 and
            minute >= 0 and minute < 60 and
            second >= 0 and second < 60)


def handle_snapshot_request(packages):
    '''Listen for user input and generate snapshot when input is valid.'''
    give_user_snapshot_instructions()

    user_requested_time = input('Whenever you\'re ready...  ')
    while not validate_length(user_requested_time):
        print('\nPlease try again. Please enter two-digit hour, colon, '
              'two-digit minute.')
        user_requested_time = input('Whenever you\'re ready...  ')

    time_parts = get_time_parts_from_string(user_requested_time)
    while not validate_values(time_parts):
        print('\nPlease try again. Please enter an hour from 0 to 23, minute '
              'from 0 to 59, and second (if entered) also from 0 to 59.')
        user_requested_time = input('Whenever you\'re ready...  ')

    print(f'Alright! You asked for a snapshot at time {user_requested_time}')
    make_snapshot(Time_Custom(*time_parts), packages)

    print('*' * 79, '\n')


def package_status_at_time(package, time_custom):
    '''Return the status of a package at a given time.'''
    # if time_custom is before 7:59am, use the initial state
    if time_custom < Time_Custom(7, 59, 00):
        return package.props['history'][0].state
    if len(package.props['history']) == 1:
        return package.props['history'][0].state
    # case: package had at least one History_Record after time_custom
    for index, record in enumerate(package.props['history']):
        if record.time > time_custom:
            # go back one record
            return package.props['history'][index - 1].state
    # case: package no History_Records after time_custom
    return package.props['history'][-1].state


def package_snapshot_file(package, time_custom):
    '''Write historical status of given package at given time to text file.'''
    filename = 'package_snapshot.txt'
    append_or_write = 'a' if path.exists(filename) else 'w'
    with open(filename, append_or_write) as f:
        package_str = str(package)

        # do not include current delivery-state of package in snapshot string
        cut_start = package_str.index('delivery status')
        cut_end = package_str.index('destination')
        print(package_str[0:cut_start] + package_str[cut_end:], file=f)

        status = package_status_at_time(package, time_custom).name
        print(f'\tFinally, delivery status at {time_custom} was {status}',
              file=f)


def package_snapshot(package, time_custom):
    '''Display historical status of a given package at a given time.'''
    package_str = str(package)

    # do not include current delivery-state of package in snapshot string
    cut_start = package_str.index('delivery status')
    cut_end = package_str.index('destination')
    print(package_str[0:cut_start] + package_str[cut_end:])

    status = package_status_at_time(package, time_custom).name
    print(f'\tFinally, delivery status at {time_custom} was {status}')


def make_snapshot(time_custom, packages):
    '''Display historical status of each/every package at a provided time.'''
    print(f'\nSNAPSHOT OF ALL PACKAGES AT {str(time_custom)}:')
    for package in packages:
        package_snapshot(package, time_custom)
        package_snapshot_file(package, time_custom)


def ask_user_if_they_have_correction_information():
    '''Ask user if they have destination-correction information.'''
    print('If you have any destination-corrections for packages, you are in '
          'the right place.\nIf you don\'t have any corrections, no problem--'
          'just type "q" or "quit" and hit Enter.\nIf you do have information, '
          'type any other key--like the Enter key.\n'
          'Just keep in mind that if your distance file indicates that any '
          'packages\nare known to have the wrong destination, but you do not '
          'supply a correction,\nthose packages will not get delivered. '
          'Remember--to exit this optional step, \njust press "q" or "quit", '
          ' then Enter, to skip it.')
    ask_user = input('')
    return ask_user.lower().strip() not in ('q', 'quit')


def give_user_correction_instructions():
    '''Give the user instructions for supplying valid destination-correction
    information.'''
    print('\tIf you just want to update a package\'s destination '
          'right away,\nfirst type the package ID, then comma and '
          'space, then either the landmark\nor street address.'
          '\n\tIf you are entering information about a destination '
          'correction that\nwill be known later, enter it in this '
          'format: first the package ID, and then\nthe time by which '
          'the true destination will be known.'
          '\n\tIf you are updating an existing correction item, '
          'enter it in this \nformat: first the package ID, and then '
          'either the landmark or the street\naddress (yes, this is '
          'the same as in the first case).'
          '\n\n\tHere are some examples of input that will be accepted:'
          '\n3, 09:30\t\t<= this says that package with ID 3 has the wrong '
          '\n\t\t\tdestination location, but it will be corrected'
          '\n\t\t\tby 9:30am today'
          '\n5, 13:30\t\t<= similarly, this says package 5 will'
          '\n\t\t\thave the correct destination by 1:30pm today'
          '\n6, Sugar House Park\t<= this says that package ID 6 needs to go '
          'to\n\t\t\tSugar House Park instead of whatever the previous '
          '\n\t\t\tdestination was'
          '\n\n\tA couple of things worth keeping in mind:'
          '\n1. If you enter a time, it must be in this format'
          '\n\th:mm or hh:mm\n\tExamples: 3:55, 12:30, 14:00'
          '\n2. If you enter a time, it must be "military time".'
          '\n3. That means "AM" and "PM" are ignored.'
          '\n4. If you enter a location, then whether that is a '
          'landmark or street address, it must match exactly what '
          'the package csv file had for that package. The exception is '
          'that the (five-digit) zip code is optional.\n')


def validate_user_correction_information(input_string, Locations):
    '''Validate user-supplied destination-correction information.

    Note the -6: This lops off the 5 zip digits plus the preceding space.
    '''
    user_input = input_string.split(',')
    if len(user_input) != 2:
        return False

    package_id, time_or_location = user_input
    package_id = package_id.strip()
    time_or_location = time_or_location.strip()

    digit_regex = re.search("(\d{1,2}):(\d{2})", time_or_location)

    landmarks = [loc.landmark for loc in Locations]
    street_addresses = [loc.address for loc in Locations]
    street_addresses_no_zip = [loc.address[:-6] for loc in Locations]

    return (user_input[0].isdigit() and   # first part an integer
            (digit_regex or               # second part a time
             time_or_location in landmarks or     # or second part a location
             time_or_location in street_addresses or
             time_or_location in street_addresses_no_zip))


def get_valid_correction_item_or_quit(Locations):
    '''Get valid destination-correction information from user.'''
    prompt = ('\nPlease enter the destination-correction information '
              'on the next line, then hit the Enter key:\n')
    sorry = ('\nSorry, that doesn\'t look like valid correction information '
             'to me.\nIf you don\'t want to enter information after all, '
             'type "q" or "quit" to exit.\nOtherwise, please try again.')

    user_input = input(prompt)
    while not validate_user_correction_information(user_input, Locations):
        print(sorry)
        user_input = input(prompt)
        if user_input.lower().strip() in ('q', 'quit', 'quit()'):
            return 'quit'

    return user_input


def get_valid_package_id_from_string(package_id_string):
    '''Validate and return package ID from string.

    This function assumes that the company's package ID numbers
    will always be less than 1000. This assumption is shared by a
    Package class method (validate_package_ID_list method).
    '''
    ID = int(package_id_string.strip())

    if not ID > 0:
        raise ValueError('Package ID found in special note too small')
    if not ID < 1000:
        raise ValueError('Package ID found in special note too large')

    return ID


def get_time_or_location_from_string(time_or_location, Locations):
    '''Extract the time or the location supplied by user for a correction;
    if user supplies both (unexpected), only the address is extracted.

    Note the -6: This lops off the 5 zip digits plus the preceding space.
    '''
    time_or_location = time_or_location.strip()

    landmarks = [loc.landmark for loc in Locations]
    if time_or_location in landmarks:
        index_of_match = landmarks.index(time_or_location)
        # this trick works because list comprehensions retain order
        return Locations[index_of_match]

    street_addresses = [loc.address for loc in Locations]
    if time_or_location in street_addresses:
        index_of_match = street_addresses.index(time_or_location)
        # this trick works because list comprehensions retain order
        return Locations[index_of_match]

    street_addresses_no_zip = [loc.address[:-6] for loc in Locations]
    if time_or_location in street_addresses_no_zip:
        index_of_match = street_addresses_no_zip.index(time_or_location)
        # this trick works because list comprehensions retain order
        return Locations[index_of_match]

    digit_regex = re.search("(\d{1,2}):(\d{2})", time_or_location)
    if digit_regex:
        hour, minute = digit_regex.groups()
        hour, minute = int(hour), int(minute)
        return Time_Custom(hour, minute, 0)


def get_one_destination_correction(Locations):
    '''Get one destination-correction from user.'''
    item_or_quit = get_valid_correction_item_or_quit(Locations)
    if item_or_quit == 'quit':
        return 'quit'

    correction_item_list = [None] * 3

    user_input = item_or_quit.split(',')

    package_id = get_valid_package_id_from_string(user_input[0])
    correction_item_list[0] = package_id

    time_or_location = get_time_or_location_from_string(
        user_input[1], Locations)
    if isinstance(time_or_location, Time_Custom):
        correction_item_list[1] = time_or_location  # time
    else:
        correction_item_list[2] = time_or_location  # location

    return correction_item_list


def get_destination_corrections(Locations):
    '''Get one or more destination-corrections from user; calls many helpers.

    Data definition:
    A Destination_Correction is a namedtuple of
        id:         Package ID
        time:       Time by which we will know the correct destination
        location:   Location* of the correct destination

    *Location is itself a namedtuple of num, landmark, address (see load.py).

    Notes on destination corrections:
        Purpose: Separating 'when it will be known' from the destination itself
    lets us handle when a sender initially gave us the wrong address (and tells
    us so), but when they need some time to find out the right address or
    otherwise get it to us. It also lets us handle when we know we got it
    wrong but know that we'll have the right information later.
        Usage: the list of destination corrections returned by this function is
    intended to be checked by truck drivers at the hub. If the time listed for
    a correction has passed, the truck is responsible for updating the package.
    If the location property of a correction is still None, despite the time
    being past the time-of-correction, the package will not be updated.
    '''
    Destination_Corrections = []

    Destination_Correction = namedtuple('Destination_Correction',
                                        ['pkg_id', 'time', 'location'])

    user_has_information = ask_user_if_they_have_correction_information()

    if not user_has_information:
        return Destination_Corrections

    give_user_correction_instructions()

    while user_has_information:
        item_or_quit = get_one_destination_correction(Locations)

        if item_or_quit == 'quit':
            break
        else:
            match = [c for c in Destination_Corrections
                     if item_or_quit[0] == c.pkg_id]
            # merge or overwrite existing correction for that package if exists
            if len(match) > 0:
                match = match[0]
                index = Destination_Corrections.index(match)
                updated = (match.pkg_id,
                           (item_or_quit[1] or match.time),
                           (item_or_quit[2] or match.location))
                Destination_Corrections[index] = (
                  Destination_Correction(*updated))
            else:  # add new correction
                Destination_Corrections.append(
                  Destination_Correction(*item_or_quit))

        ask_for_more = input('\nWould you like to add another? Type "y" '
                             'or "yes" if so.\nOtherwise, hit Enter.\n')
        user_has_information = ask_for_more.lower().strip() in ('y', 'yes')

    print('*' * 79, '\n')

    return Destination_Corrections



from .specific_tests.algorithms_tests import test_algorithms
from .specific_tests.hash_tests import test_hashes
from .specific_tests.regex_tests import test_regexes


def test():
    test_algorithms()
    test_hashes()
    test_regexes()



from itertools import permutations
from .time_custom import Time_Custom


class ImproveRoute_Min_ValueError(BaseException):
    pass


def update_subroute_distances(subroute, distances):
    '''Update a subroute to have correct distances.

    Receives subroute as list of tuples in which each tuple's indices are
        0 = location-number     1 = distance-from-previous  2 = package-list
    We are updating the [1] index, distance-from-previous.
    '''
    for index, stop in enumerate(subroute):
        if index == 0:
            continue
        prev, curr = subroute[index-1][0], stop[0]
        dist = distances[prev][curr]
        subroute[index] = stop[0], dist, stop[2]
    return subroute


def meets_deadlines(partial_route, distances, deadlines, speed, leave_time):
    '''Return whether a given partial-route meets all package deadlines.

    NOTE: I override speed with 18 (see below).
    The requirements for which this project built specify that the average
    speed of a truck is always 18mph, no matter what, and including stops.
    I built other parts of this program to be flexible via accommodating an
    actual speed function in the future. However, I did not deem it worth
    the effort to make this particular function flexible wrt a speed function.
    '''
    if deadlines == []:
        return True

    speed = 18

    distance_so_far = 0

    for stop in partial_route:
        distance_so_far += stop[1]  # stop[1] is distance-from-previous-stop

        # stop[0] and d[0] are both location-numbers
        if stop[0] in [d[0] for d in deadlines]:
            minutes_to_get_there = 60 * (distance_so_far / speed)
            projected_arrival = Time_Custom.clone(leave_time)
            projected_arrival.add_time(minutes_to_get_there)

            deadline, = [d[1] for d in deadlines if stop[0] == d[0]]
            if projected_arrival > deadline:
                return False

    return True


def recreate_namedtuples(route, Stop_namedtuple):
    '''Recreate namedtuples (permutations downcasts to regular tuples).'''
    return [Stop_namedtuple(*stop_tuple)
            if not isinstance(stop_tuple, Stop_namedtuple)
            else stop_tuple
            for stop_tuple in route]


def improve_route(route, distances, deadlines, speed, leave, Stop_namedtuple):
    '''Reorder the ordering of stops in segments (or subroutes) of size 7
    whenever a shorter segment distance can be found by reordering.

    Why 7? Please read on. (tl;dr to balance runtime and optimality).
    * For a segment of size n, check every possible permutation of
    stop-orders to find the one with the shortest distance.
    * If we did this for the entire route, we'd have a guaranteed shortest
    distance--but the runtime would be O(n!)
    * However.. O(n!) is not so bad if we limit n. I know--this seems TOO
    easy/simple. But hey--why NOT check if subroutes can be reordered to
    make a shorter overall route?
    * For a given segment, I keep the start and end fixed and permute the
    order of stops in between. This means (n-2)! orderings are checked per
    segment. This means (m-n) * (n-2)! total permutations are checked,
    where m = total route length.
    * For n=7, (n-2)! = 5! = 120, which is not so bad. It's very fast to
    compute one route distance and computing 120 isn't so bad either.
    '''
    if len(route) <= 3:
        return route

    n = 7

    # Why "- n + 1?" Example: a route of size n+1 has two subroutes of size n
    for index in range(len(route) - n + 1):
        end = min(index + n - 1, len(route) - 1)  # route size can be < n

        # note that 'end' is used, not end-1, because slice-ends are exclusive
        new_orderings = list(permutations(route[index+1:end]))

        # a full subroute for distance calculation must include start and end
        subroutes = [[route[index], *ordering, route[end]]
                     for ordering in new_orderings]

        with_updated_distances = [update_subroute_distances(
                                    subroute, distances)
                                  for subroute in subroutes]

        with_none_late = [subroute for subroute in with_updated_distances
                          if meets_deadlines(route[:index] + list(subroute),
                                             distances, deadlines,
                                             speed, leave)]

        with_distance_sums = [(subroute, sum([x[1] for x in subroute]))
                              for subroute in with_none_late]

        try:
            shortest = min(with_distance_sums,
                           key=lambda candidate_route: candidate_route[1])
        except ValueError:  # probably min() was given an empty list
            raise ImproveRoute_Min_ValueError('No route exists that would '
                                              'meet all remaining deadlines.')

        route = route[:index] + list(shortest[0]) + route[end+1:]

    return recreate_namedtuples(route, Stop_namedtuple)



import re


class Time_Custom:
    '''Time objects represent a truck's clock. Minutes can be added to them.

    Objects are military time (no AM/PM) and do not have fractional seconds.
    Example usage:
        given:  Time_Custom(13,15,27)
        add:    33.5 minutes
        result: Time_Custom(13,48,57)
    '''

    digit_regex = re.compile("(\d{1,2}):(\d{2}) (am|pm)", re.IGNORECASE)

    def __init__(self, hour, minute, second):
        '''Create Time_Custom object.'''
        self.hour = hour
        self.minute = minute
        self.second = second

    def add_time(self, minutes):
        '''Advance the "clock" of a Time_Custom object by adding minutes.

        Input is restricted to a (positive number of) minutes.

        This function assumes only that the travel time between any two
        destinations does not exceed 16 hours.
        '''
        if minutes < 1 or minutes > 960:
            raise ValueError('Ineligible number of minutes passed')
        else:
            h, m, s = Time_Custom.decompose(minutes)
            h, m, s = (h + self.hour), (m + self.minute), (s + self.second)
            m, s = Time_Custom.adjust_seconds(m, s)
            h, m = Time_Custom.adjust_minutes(h, m)
            self.hour, self.minute, self.second = h, m, s

    @staticmethod
    def decompose(minutes_to_add):
        '''Break down minutes-to-add into hours, minutes and seconds.'''
        h = int(minutes_to_add / 60)
        m = int(minutes_to_add - 60 * h)
        s = round(60 * (minutes_to_add % 1))
        return h, m, s

    @staticmethod
    def adjust_minutes(h, m):
        '''Adjust any number of seconds > 60 by rolling over into minutes.'''
        hours_overflow = int((m - (m % 60)) / 60)
        return h + hours_overflow, m - 60 * hours_overflow

    @staticmethod
    def adjust_seconds(m, s):
        '''Adjust any number of minutes > 60 by rolling over into hours.'''
        minutes_overflow = int((s - (s % 60)) / 60)
        return m + minutes_overflow, s - 60 * minutes_overflow

    @classmethod
    def clone(cls, time_obj):
        '''Return a clone of a time-object.'''
        return Time_Custom(time_obj.hour, time_obj.minute, time_obj.second)

    @classmethod
    def is_valid_AM_PM_time(cls, time_string):
        '''Return whether time_string can be parsed as a valid Time_Custom.'''
        time_parts = cls.digit_regex.search(time_string)

        if time_parts is None or len(time_parts.groups()) != 3:
            return False

        hour, minute, am_pm = time_parts.groups()
        hour, minute = int(hour), int(minute)

        if not ((hour >= 0 and hour <= 12) and
                (minute >= 0 and minute < 60) and
                not (hour == 0 and am_pm in ('pm', 'PM'))):
            return False

        return True

    @classmethod
    def make_time_from_string(cls, time_string):
        '''Make Time_Custom object from string.'''
        time_parts = cls.digit_regex.search(time_string)
        hour, minute, am_pm = time_parts.groups()
        hour, minute = int(hour), int(minute)

        if am_pm == 'pm':
            hour += 12

        deadline_tuple = (hour, minute, 0)
        return Time_Custom(*deadline_tuple)

    def __str__(self):
        '''Return string representation of a Time_Custom object.'''
        return f'{self.hour:02}:{self.minute:02}:{self.second:02}'

    def __eq__(self, other):
        '''Return whether one time is == another. Assumes inputs are times.'''
        if not isinstance(other, Time_Custom):
            raise ValueError('Right-hand side of == not a Time_Custom object')
        elif (self.hour == other.hour and
              self.minute == other.minute and
              self.second == other.second):
            return True
        else:
            return False

    def __ne__(self, other):
        '''Return whether one time is != another. Assumes inputs are times.'''
        return not self == other

    def __lt__(self, other):
        '''Return whether one time is < another. Assumes inputs are times.'''
        if not isinstance(other, Time_Custom):
            raise ValueError('Right-hand side of < not a Time_Custom object')
        else:
            if self.hour < other.hour:
                return True
            elif (self.hour == other.hour and
                  self.minute < other.minute):
                return True
            elif (self.minute == other.minute and
                  self.second < other.second):
                return True
            else:
                return False

    def __gt__(self, other):
        '''Return whether one time is > another. Assumes inputs are times.'''
        if not isinstance(other, Time_Custom):
            raise ValueError('Right-hand side of > not a Time_Custom object')
        else:
            if self.hour > other.hour:
                return True
            elif (self.hour == other.hour and
                  self.minute > other.minute):
                return True
            elif (self.minute == other.minute and
                  self.second > other.second):
                return True
            else:
                return False

    def __le__(self, other):
        '''Return whether one time is <= another. Assumes inputs are times.'''
        return self == other or self < other

    def __ge__(self, other):
        '''Return whether one time is >= another. Assumes inputs are times.'''
        return self == other or self > other



import csv
import re
from collections import namedtuple
from .classes.package import Package
from .classes.time_custom import Time_Custom


class DistanceCsv_ValueError(BaseException):
    pass


class PackageCsv_ValueError(BaseException):
    pass


def read_distance_csv(csv_file):
    '''Return list-of-lists representation of the contents of passed-in csv.

    Remarks on csv: This whole program expects the passed-in csv to contain a
    (skippable) header. It expects the list of locations appearing in the
    first column to be identical to the list of locations in the first row.
    It expects there to be a second column with location data, but this column
    contains only street-addresses and zip-codes, not landmark names (like the
    first row and column have). Finally, it expects that zip codes will
    consistently appear only in this second column, and they will be wrapped
    in parentheses.
    None of these expectations should be read as an endorsement of the format
    of the input csv. This is simply the format that the user provides.
    '''
    csv_data = []
    with open(csv_file) as csvFile:
        csvReader = csv.reader(csvFile)
        for row in csvReader:
            if row[0] != '':
                csv_data.append(row)
    return csv_data


def clean_address_data(column_one, column_two):
    '''Return tuple of cleaned landmark and street address.'''

    # In column one of csv, street-address is on a newline after landmark name
    landmark = column_one.split('\n')[0].strip()

    # In column two the format is 'address\n(zip)'--I prefer 'address zip'
    # and that is the format to which package addresess will be matched
    street_address = column_two.replace('\n(', ' ').rstrip(')').strip()

    # Package addresses will have the same standard for direction abbreviation
    # so that the two sets of addresses are consistent and therefore matchable
    street_address = (street_address
                      .replace('South', 'S')
                      .replace('North', 'N')
                      .replace('East', 'E')
                      .replace('West', 'W'))

    return landmark, street_address


def get_location_data(csv_data):
    '''Return list of tuples of index, landmarks, and street addresses.'''
    locations = []
    for index, row in enumerate(csv_data):
        if index != 0:
            landmark, street_address = clean_address_data(row[0], row[1])
            locations.append((index, landmark, street_address))
    return locations


def populate_locations(location_data, Location_constructor):
    '''Return list of location named-tuples from location data.'''
    Locations = []
    for location_datum in location_data:
        Locations.append(Location_constructor(*location_datum))
    return Locations


def clean_distance_data(csv_data):
    '''Remove redundant column, use location numbers instead of names,
    and convert distance data (a mix of integer and string) to floats.
    '''
    for row_index, row in enumerate(csv_data):
        del row[1]

        for col_index, datum in enumerate(row):
            if row_index != 0 and col_index != 0:
                if datum != '':
                    csv_data[row_index][col_index] = float(datum)

        # row_index will be used for location numbers
        if row_index != 0:
            row[0] = row_index

    # col_index will be used for location numbers
    # Note that col_index coincides with (is the same value as) row_index.
    for col_index, _ in enumerate(csv_data[0]):
        if col_index != 0:
            csv_data[0][col_index] = col_index


def fill_distance_data(data):
    '''Fill in distance data where the missing pieces of data can be inferred.

    If cell at [i,j] is missing, but [j,i] is known, fill [i,j] with [j,i],
    and vice versa. Skip when neither or both are empty.
    '''
    for row_index, row in enumerate(data):
        for col_index, distance in enumerate(row):
            ij = data[row_index][col_index]
            ji = data[col_index][row_index]

            if ij == '' and ji != '':
                data[row_index][col_index] = data[col_index][row_index]

            elif ji == '' and ij != '':
                data[col_index][row_index] = data[row_index][col_index]


def validate_distance_data(data):
    '''Validate distance data is both present and non-contradicting.'''
    for row_index, row in enumerate(data):
        for col_index, distance in enumerate(row):
            ij = data[row_index][col_index]
            ji = data[col_index][row_index]

            if ij is None and ji is None:
                return False
            if ij != ji:
                return False

    return True


def read_package_csv(csv_file):
    '''Return list-of-lists representation of the contents of passed-in csv.

    Remarks on csv: This whole program expects the passed-in csv to contain
    a list of package attributes in the following order:
        - Package ID
        - Address
        - City
        - State
        - Zip
        - Delivery Deadline (a time-of-day)
        - Mass (weight in kg)
        - Special Notes
            Note: The plural is a misnomor--only one special note is permitted
    '''
    csv_data = []
    with open(csv_file) as csvFile:
        csvReader = csv.reader(csvFile)
        for row in csvReader:
            if row[0].isdigit():
                csv_data.append(row)
    return csv_data


def clean_package_data(csv_data):
    '''Clean package data preliminarily to initializing package objects.'''
    for row in csv_data:
        # if deadline is 'EOD' or end of day, it can be ignored
        if row[5] == 'EOD':
            row[5] = None

        # Distance addresses have the same standard for direction abbreviation
        # so that the two sets of addresses are consistent and thus matchable
        row[1] = (row[1]
                  .replace('South', 'S')
                  .replace('North', 'N')
                  .replace('East', 'E')
                  .replace('West', 'W'))

        # only the first eight columns have data
        del row[8:]


def validate_package_address_data(package_data, location_namedtuples):
    '''Validate all package address data matches a location address.'''
    location_addresses = [loc[2] for loc in location_namedtuples]
    for row in package_data:
        package_address = f'{row[1]} {row[4]}'  # street address plus zip code
        if package_address not in location_addresses:
            return False
    return True


def get_one_package_destination(pkg_row, location_namedtuples):
    '''Return the location named-tuple matching a package's destination.'''
    package_address = f'{pkg_row[1]} {pkg_row[4]}'  # street address plus zip
    location_addresses = [loc[2] for loc in location_namedtuples]
    return location_namedtuples[location_addresses.index(package_address)]


def validate_package_deadline_times(package_data):
    '''Validate that all package deadlines can be converted to Time_Custom.'''
    for row in package_data:
        deadline = row[5]
        if deadline is None:
            continue

        if not Time_Custom.is_valid_AM_PM_time(deadline):
            return False

    return True


def populate_packages(package_data, location_namedtuples):
    '''Return list of Package objects based on package data from the csv.'''
    all_packages = []
    for package_row in package_data:
        destination = get_one_package_destination(package_row,
                                                  location_namedtuples)
        deadline = package_row[5]
        if deadline is not None:
            deadline = Time_Custom.make_time_from_string(package_row[5])

        data_to_use = [package_row[0], deadline] + package_row[6:8]
        new_package = Package(*data_to_use, destination)
        all_packages.append(new_package)
    return all_packages


def load_data(distance_csv, package_csv):
    '''Populate packages list, distances 2D list, Locations namedtuple list.

    Data definition:
    A Location is a namedtuple of location-number, landmark, street address.
    '''
    distances, Locations, packages = [], [], []

    Location = namedtuple('Location', ['num', 'landmark', 'address'])

    distances = read_distance_csv(distance_csv)

    # Locations must be populated before clean_distance_data removes addresses
    location_data = get_location_data(distances)
    Locations = populate_locations(location_data, Location)
    clean_distance_data(distances)
    fill_distance_data(distances)

    if not validate_distance_data(distances):
        raise DistanceCsv_ValueError('One or more distance values are absent '
                                     'or contradict other values in the file')

    package_data = read_package_csv(package_csv)
    clean_package_data(package_data)

    if not validate_package_address_data(package_data, Locations):
        raise PackageCsv_ValueError('One or more package addresses does not '
                                    'match any location from distances csv')
    if not validate_package_deadline_times(package_data):
        raise PackageCsv_ValueError('One or more package deadlines could not '
                                    'be parsed as a valid AM/PM time')

    packages = populate_packages(package_data, Locations)

    return distances, Locations, packages



from ...classes.hash import *


def test_hashes():
    # 1 PASS
    # initialize a Hash with two values and print the hash. This will
    hash_one = Hash('bob', 'joe')
    # print("I expect to see a list of 48 Nones and "
    # "two key-value lists of form [value, None]")
    # print(hash_one._debug_str())

    # GET TESTS
    # 2 PASS
    # Key not found, test 1: None is found at the index
    # print("I expect hash_one['sarah'] to raise a KeyError")
    # print(hash_one['sarah'])

    # 3 PASS
    # Key not found, test 2: a key-value list is found at the index,
    # but the key does not match the key being requested
    # print("I expect hash_one['cnb'] to raise a KeyError")
    # print(hash_one['cnb'])

    # 4 PASS
    # Key not found, test 3: a list of key-value lists is found at the index,
    # but none of the keys in the key-value lists match the requested key
    # hash_one['cnb'] = 55  # creating a list of key-value lists at that index
    # print("I expect hash_one['dmb'] to raise a KeyError")
    # print(hash_one['dmb'])

    # 5 PASS
    # Get from a single key-value list
    # In the 'bob' case below, I expect the default value to be None
    # This situation will occur when a new Hash is initialized with arguments.
    # hash_one['new_key'] = 42
    # print("I expect hash_one['new_key'] to return 42")
    # print(hash_one['new_key'])
    # print("I expect hash_one['bob'] to return None")
    # print(hash_one['bob'])

    # 6 PASS
    # NOTE: this is when my hashing algorithm was weaker. I could find it
    # by searching my git commits if need be. Back then, bob=cnb
    # Get from a list of key-value lists
    # hash_one['cnb'] = 55
    # print(f"Showing there is a list of lists now: {hash_one._debug_str()}")
    # print("I expect hash_one['bob'] to return None")
    # print(hash_one['bob'])

    # SET TESTS
    # 7 PASS
    # assign to an index currently holding None
    # this should replace None with [key, value]
    # NOTE: this is the case that the vast majority of hashes should fall into
    # hash_one['dork'] = 30
    # print("I expect hash_one['dork'] to return 30")
    # print(hash_one['dork'])

    # 8 PASS
    # reassign an existing key
    # (no hash collision, i.e., only one key-value list is at that index)
    # hash_one['bob'] = 7
    # print("I expect hash_one['bob'] to return 7")
    # print(hash_one['bob'])

    # 9 PASS
    # NOTE: this is when my hashing algorithm was weaker. I could find it
    # by searching my git commits if need be. Back then, bob=cnb
    # hash collision case 1: one key-value list exists at that index already
    # but the key living there is not the key being inserted
    # this should wrap the existing key-value list in a list,
    # and then append to the outer list
    # hash_one['cnb'] = 55
    # print(f"Showing there is a list of lists now: {hash_one._debug_str()}")

    # 10 PASS
    # hash collision case 2: multiple key-value lists exist there already
    # and the key being updated already exists in that list
    # in this case, reassign the existing key
    # hash_one['cnb'] = 55
    # print("I expect hash_one['cnb'] to return 55")
    # print(hash_one['cnb'])
    # hash_one['cnb'] = 37
    # print("I expect hash_one['cnb'] to return 37")
    # print(hash_one['cnb'])

    # 11 PASS
    # hash collision case 3: multiple key-value lists exist there already
    # and the key in question does not yet exist among the list
    # in this case, append to the existing list of key-value lists
    # hash_one['cnb'] = 55
    # hash_one['dmb'] = 44
    # print("I expect hash_one['dmb'] to return 44")
    # print(hash_one['dmb'])

    # TESTS ADDED LATER
    # 12 PASS
    # initialize Hash object with key-value lists
    # hash_two = Hash(['hello', 4], ['world', 5])
    # print("I expect to see [hello,4] and [world,5] in hash_two")
    # print(hash_two._debug_str())

    # 13 PASS
    # initialize Hash object with a mix of key-strings and key-value tuples
    # hash_three = Hash('hello', ('world', 5))
    # print("I expect to see [hello,None] and [world,5] in hash_two")
    # print(hash_three._debug_str())

    # 14 PASS
    # as soon as 60% or more of buckets/slots are filled,
    # hash_size (buckets/slots) is doubled
    # hash_four = Hash(*[c for c in 'abcdefghijklmnopqrstuvwxyz0123'])
    # print("I expect there to be 50 slots")
    # print('length: ', len(hash_four._props))
    # hash_four['%'] = 8
    # print("I expect there to be 100 slots")
    # print('length: ', len(hash_four._props))
    # print(hash_four._debug_str())

    # 15 PASS
    # as soon as 60% or more of buckets/slots are filled, all keys are
    # re-hashed. NOTE: nested lists count once per sublist.
    # hash_five = Hash(*[c for c in 'abcdefghijklmnopqrstuvwxyz0123'])
    # print("I expect the ordering of items to be different between 1 and 2.")
    # print("len(1)", len(hash_five._props), " and 1:", hash_five._debug_str())
    # hash_five['%'] = 8
    # print("len(2)", len(hash_five._props), " and 2:", hash_five._debug_str())

    # 16 PASS
    # NOTE: this is when my hashing algorithm was weaker. I could find it
    # by searching my git commits if need be. Back then, bob=cnb
    # str method correctly formats hashes with sublists and excludes all Nones
    # hash_six = Hash('bob', 'cnb')  # bob and cnb hash to same index
    # print(str(hash_six))
    # hash_six_point_one = Hash('bob')
    # print(str(hash_six_point_one))
    # hash_six_point_two = Hash()
    # print(str(hash_six_point_two))

    # 17 PASS
    # getitem raises an error if not given a string
    # print("I expect to see a TypeError with hash_six[4]")
    # hash_six[4]

    # 18 PASS
    # setitem raises an error if not given a string
    # print("I expect to see a TypeError with hash_six[4] = 6")
    # hash_six[4] = 6

    # 19
    # str method does not break if a hashed value equals itself
    # hash_seven = Hash('hi','world')
    # hash_seven['recursion!'] = hash_seven
    # print("I expect to see..")
    # print(str(hash_seven))

    # # 20 PASS
    # # _debug_str method represents props as what they are--nested lists
    # hash_eight = Hash(('hey', 4))
    # print(hash_eight._debug_str())

    # 21 PASS
    # deepcopy method works
    # b = [[[3,4],[5],6],[7,[8,9]],10]
    # b_ = Hash._deepcopy(b)
    # print(b)
    # print(b_)
    # print('This should fail: id b_ == id b:', id(b_)==id(b))

    # 22 PASS
    # initializing Hash with None throws KeyError (because None not a string)
    # hash_nine = Hash(None)
    # print("I expect to see a KeyError")
    # print(hash_nine._debug_str())

    # 23 PASS
    # hash_ten, hash_eleven = Hash(['hello', 5]), Hash()
    # print("Before adding prop to hash_eleven I expect False for == test")
    # print(hash_ten == hash_eleven)
    # print("After adding prop to hash_eleven I expect True for == test")
    # hash_eleven['hello'] = 5
    # print(hash_ten == hash_eleven)

    # 24 PASS
    # hash_twelve = Hash(['world', 5])
    # print("I expect get to behave the same as [] when a value has been set")
    # print(f"Expecting 5: {hash_twelve.get('world')}")
    # print("I expect get to supply a default value if no value has been set")
    # print(f"Expecting None: {hash_twelve.get('hello')}")
    # print("I expect get to use a passed-in default value")
    # print(f"Expecting 'not found!': {hash_twelve.get('hello','not found!')}")

    # 25 PASS
    # eval(repr(obj) should == obj
    # hash_13 = Hash('hey', ['yo', 5])
    # print("I expect this to be True: hash_13 == eval(repr(hash_13))")
    # print(hash_13 == eval(repr(hash_13)))

    # 26 PASS
    # passing kwargs into Hash() adds them
    # hash_14 = Hash(a_kwarg=45)
    # print("I expect to see a Hash with k-v pair [a_kwarg, 45]")
    # print(hash_14._debug_str())



class Hash():
    '''The Hash class provides hash-table objects.

    Public API / Examples of Usage, in Ten Lines:
        example = Hash()  # initialize
        example['hello'] = 5
        example['hello']  # returns 5
        str(example)  # returns { hello: 5 }  (but {} on separate lines)
        example['hey']  # raises KeyError
        example2 = Hash(3)  # raises TypeError (keys must be strings)
        example3 = Hash('hi')  # adds a key of 'hi' and default value of None
        example4 = Hash(['hi', 3])  # adds a key of 'hi' with value 3
        example5 = Hash(('hi', 6))  # adds a key of 'hi' with value 6
        example6 = Hash(a_kwarg=45)  # adds a key of 'a_kwarg' with value 45

    Attributes (Instance variables):
     - hash_size: all Hash objects start with 50 buckets/slots
     - props: this stores the hashed values. Initialized with Nones.

    Note on hash collision: the chaining approach is used.
      If multiple values have the same hashed index, there will be a list of
      key-value lists at that index.
      For the more common case of no collision for a given index, there will be
      only a single key-value list at an index.

    Further notes:
      - Under the hood, hashed keys and their values are stored as nested lists
      - Hash objects are doubled and rehashed when # items >60% of __props size
      - The internal _hash method is basic and should not be regarded as secure
    '''

    def __init__(self, *args, **kwargs):
        '''Create Hash object.'''
        self._hash_size = 50
        self._props = [None] * self._hash_size
        self._count = 0

        for k, v in kwargs.items():
            self[k] = v

        for arg in args:
            if isinstance(arg, (list, tuple)) and len(arg) == 2:
                key_string, value = arg[0], arg[1]
                self[key_string] = value
            else:
                self[arg] = None

    def get(self, key_string, default=None):
        '''Return property of hash object or default value if key not found.

        This is identical to __getitem__ except "raise KeyError.." is replaced
        with "return default".
        '''
        if not isinstance(key_string, str):
            raise TypeError(f'Key {key_string} is not a string')

        index = self._hash(key_string)
        prop = self._props[index]

        # case: None was at that index
        if prop is None:
            return default

        # case: a list of key-value lists was at that index
        if isinstance(prop[0], list):
            for sub_list in prop:
                if sub_list[0] == key_string:
                    return sub_list[1]
            return default

        # case: a single key-value list was at that index
        if prop[0] == key_string:
            return prop[1]
        else:
            return default

    def _flatten(self):
        '''Flatten props. This method exists to be called by __repr__.'''
        flat = []

        for prop in self._props:
            if prop is None:
                continue

            if isinstance(prop[0], list):
                for sub_list in prop:
                    key_string, value = sub_list[0], sub_list[1]
                    flat.append([key_string, value])

            else:
                key_string, value = prop[0], prop[1]
                flat.append([key_string, value])

        return flat

    def _hash(self, key_string, str_size_min=10):
        '''Hash a key and return index.'''
        padded = key_string.rjust(str_size_min, key_string[0])
        return sum([(ord(ch)) ** ((idx % 30) + 5) + 43
                   for idx, ch in enumerate(padded)]
                   ) % self._hash_size

    def _rehash(self):
        '''Re-hash all items (called when self._props needs to double).'''
        copy = self._double()
        self._count = 0

        for prop in copy:
            if prop is None:
                continue

            if isinstance(prop[0], list):
                for sub_list in prop:
                    key_string, value = sub_list[0], sub_list[1]
                    self[key_string] = value

            else:
                key_string, value = prop[0], prop[1]
                self[key_string] = value

    def _double(self):
        '''Double hash object props size and re-hash all keys.

        The disregard of the copy module's deepcopy method is intentional,
        in order to keep this class free of imports.
        '''
        copy = Hash._deepcopy(self._props)

        self._hash_size *= 2
        self._props = [None] * self._hash_size

        return copy

    def _debug_str(self):
        '''Return string representation of hash object useful for debugging.

        Whereas str prints self._props as if the underlying data structure
        were (itself) a hash, debug_str shows what Hash objects truly are:
        nested lists.
        '''
        return f"\nprops: {','.join([str(prop) for prop in self._props])}\n"

    def __getitem__(self, key_string):
        '''Get property of hash object via [] notation.'''
        if not isinstance(key_string, str):
            raise TypeError(f'Key {key_string} is not a string')

        index = self._hash(key_string)
        prop = self._props[index]

        # case: None was at that index
        if prop is None:
            raise KeyError(f'Key {key_string} does not exist in the hash')

        # case: a list of key-value lists was at that index
        if isinstance(prop[0], list):
            for sub_list in prop:
                if sub_list[0] == key_string:
                    return sub_list[1]
            raise KeyError(f'Key {key_string} does not exist in the hash')

        # case: a single key-value list was at that index
        if prop[0] == key_string:
            return prop[1]
        else:
            raise KeyError(f'Key {key_string} does not exist in the hash')

    def __setitem__(self, key_string, value):
        '''Set property of hash object via [] notation.'''
        if not isinstance(key_string, str):
            raise TypeError(f'Key {key_string} is not a string')

        index = self._hash(key_string)
        prop = self._props[index]

        # case: None was at that index
        if prop is None:
            self._props[index] = [key_string, value]
            self._count += 1

        # case: a list of key-value lists was at that index
        elif isinstance(prop[0], list):
            for sub_list in prop:
                if sub_list[0] == key_string:
                    sub_list[1] = value
            # if not found:
            self._props[index].append([key_string, value])
            self._count += 1

        # case: a single key-value list was at that index
        # note: at this point, isinstance(prop, list) must be True
        else:
            if prop[0] == key_string:
                self._props[index][1] = value
            else:  # if not found, wrap existing item in a list and append
                current_resident = self._props[index]
                self._props[index] = [current_resident]
                self._props[index].append([key_string, value])
                self._count += 1

        # rehash if self._props is getting too full
        if self._count / self._hash_size >= 0.6:
            self._rehash()

    def __str__(self):
        '''Return string representation of hash object.'''
        prop_strings = []

        for prop in self._props:
            if prop is None:
                continue

            if isinstance(prop[0], list):
                props_from_subl = []

                for sub_l in prop:
                    # avoid infinite recursion if a value happens to be 'self'
                    if sub_l[1] is self:
                        props_from_subl.append(f'\n\t\t{repr(sub_l[0])}: '
                                               '"self"')
                    else:
                        props_from_subl.append(f'\n\t\t{repr(sub_l[0])}: '
                                               f'{repr(sub_l[1])}')

                prop_strings.append(
                  '\n\t{' + ','.join(props_from_subl) + '\n\t}')

            else:
                # avoid infinite recursion if a value happens to be 'self'
                if prop[1] is self:
                    prop_strings.append(f'\n\t{repr(prop[0])}: "self"')
                else:
                    prop_strings.append(f'\n\t{repr(prop[0])}: '
                                        f'{repr(prop[1])}')

        return '{' + ','.join(prop_strings) + '\n}'

    def __repr__(self):
        '''Return string representation of hash object.'''
        unpacked = ', '.join(str(x) for x in self._flatten())
        return f'Hash({unpacked})'

    def __eq__(self, other):
        '''Compare two hash objects to check if they are equal.

        Two hash objects are equal if they have the same (non-None) props.
        '''
        return self._props == other._props

    @staticmethod
    def _deepcopy(lst):
        '''Create a deep copy of a list.

        This method was created intentionally to cut out an unneeded import,
        namely deepcopy from copy.
        '''
        result = []

        for item in lst:
            if not isinstance(item, list):
                result.append(item)

            else:
                result.append(Hash._deepcopy(item))

        return result
